import pprint
import logging
from bravado_core.spec import Spec
from bravado_core.operation import Operation
from bravado_core.validate import validate_schema_object
from pymacaron_core.exceptions import ValidationError
from pymacaron_core.models import generate_model_class
from pymacaron_core.models import get_model


log = logging.getLogger(__name__)


class EndpointData():
    """Just holding some info about an api endpoint"""
    path = None
    method = None
    handler_server = None
    handler_client = None
    decorate_server = None
    decorate_request = None
    operation = None
    produces_json = False
    produces_html = False

    param_in_body = False
    param_in_query = False
    param_in_path = False
    param_in_formdata = False

    no_params = False

    def __init__(self, path, method):
        self.path = path
        self.method = method.upper()

class ApiSpec():
    """Object holding the swagger spec as a YAML dict and a bravado-core Spec object,
    as well as methods for exploring the spec.
    """
    swagger_dict = None
    spec = None
    definitions = None

    host = None
    port = None
    protocol = None
    version = None
    verify_ssl = True

    def __init__(self, swagger_dict, formats=None, host=None, port=None, proto=None, verify_ssl=True):

        self.swagger_dict = swagger_dict

        config = {
            'validate_responses': True,
            'validate_requests': True,
            'validate_swagger_spec': False,
            'use_models': True,
        }

        if formats:
            assert type(formats).__name__ == 'list'
            config['formats'] = formats

        self.spec = Spec.from_dict(self.swagger_dict, config=config)
        self.definitions = self.spec.definitions

        self.host = swagger_dict.get('host', None)
        if not self.host:
            raise Exception("Swagger file has no 'host' entry")
        if host:
            self.host = host

        schemes = swagger_dict.get('schemes', None)
        if 'https' in schemes:
            self.port = 443
            self.protocol = 'https'
        elif 'http' in schemes:
            self.port = 80
            self.protocol = 'http'
        else:
            raise Exception("Swagger schemes contain neither http nor https: %s" % pprint.pformat(schemes))

        if port:
            self.port = port

        if proto:
            self.protocol = proto

        if not verify_ssl:
            self.verify_ssl = False

        self.version = swagger_dict.get('info', {}).get('version', '')


    def load_models(self, do_persist=True):
        """Generate PyMacaron Model classes for every data model in that API and store
        them in the calling api object"""

        names = []
        for model_name in self.definitions:

            model_spec = self.swagger_dict['definitions'][model_name]

            # Should this model inherit from a base class?
            parent_name = None
            if 'x-parent' in model_spec:
                parent_name = model_spec['x-parent']

            # Is this model persistent?
            persist = None
            if do_persist and 'x-persist' in model_spec:
                persist = model_spec['x-persist']

            # Associate model generator to ApiPool().<api_name>.model.<model_name>
            log.debug("Generating model class for %s" % model_name)
            generate_model_class(
                name=model_name,
                bravado_class=self.definitions.get(model_name),
                swagger_dict=model_spec,
                swagger_spec=self.spec,
                parent_name=parent_name,
                persist=persist,
                properties=model_spec['properties'] if 'properties' in model_spec else {},
            )

            names.append(model_name)

        return names


    def model_to_json(self, object, cleanup=True):
        """Take a model instance and return it as a json struct"""
        return object.to_json()


    def json_to_model(self, model_name, j, keep_datetime=False):
        """Take a json struct and a model name, and return a model instance"""
        cls = get_model(model_name, keep_datetime=keep_datetime)
        return cls.from_json(j)


    def validate(self, model_name, object):
        """Validate an object against its swagger model"""
        if model_name not in self.swagger_dict['definitions']:
            raise ValidationError("Swagger spec has no definition for model %s" % model_name)
        model_def = self.swagger_dict['definitions'][model_name]
        log.debug("Validating %s" % model_name)
        return validate_schema_object(self.spec, model_def, object)


    def call_on_each_endpoint(self, callback):
        """Find all server endpoints defined in the swagger spec and calls 'callback' for each,
        with an instance of EndpointData as argument.
        """

        if 'paths' not in self.swagger_dict:
            return

        for path, d in list(self.swagger_dict['paths'].items()):
            for method, op_spec in list(d.items()):
                data = EndpointData(path, method)

                # Which server method handles this endpoint?
                if 'x-bind-server' not in op_spec and 'operationId' not in op_spec:
                    if 'x-no-bind-server' in op_spec:
                        # That route should not be auto-generated
                        log.info("Skipping generation of %s %s" % (method, path))
                        continue
                    else:
                        raise Exception("Swagger api defines no x-bind-server for %s %s" % (method, path))

                if 'operationId' in op_spec:
                    data.handler_server = op_spec['operationId']
                else:
                    data.handler_server = op_spec['x-bind-server']

                # Make sure that endpoint only produces 'application/json'
                if 'produces' not in op_spec:
                    raise Exception("Swagger api has no 'produces' section for %s %s" % (method, path))
                if len(op_spec['produces']) != 1:
                    raise Exception("Expecting only one type under 'produces' for %s %s" % (method, path))
                if op_spec['produces'][0] == 'application/json':
                    data.produces_json = True
                elif op_spec['produces'][0] == 'text/html':
                    data.produces_html = True
                else:
                    raise Exception("Only 'application/json' or 'text/html' are supported. See %s %s" % (method, path))

                # Which client method handles this endpoint?
                if 'x-bind-client' in op_spec:
                    data.handler_client = op_spec['x-bind-client']

                # Should we decorate the server handler?
                if 'x-decorate-server' in op_spec:
                    data.decorate_server = op_spec['x-decorate-server']

                # Should we manipulate the requests parameters?
                if 'x-decorate-request' in op_spec:
                    data.decorate_request = op_spec['x-decorate-request']

                # Generate a bravado-core operation object
                data.operation = Operation.from_spec(self.spec, path, method, op_spec)

                # Figure out how parameters are passed: one json in body? one or
                # more values in query?
                if 'parameters' in op_spec:
                    params = op_spec['parameters']
                    for p in params:
                        if p['in'] == 'body':
                            data.param_in_body = True
                        if p['in'] == 'query':
                            data.param_in_query = True
                        if p['in'] == 'path':
                            data.param_in_path = True
                        if p['in'] == 'formData':
                            data.param_in_formdata = True

                    if data.param_in_path:
                        # Substitute {...} with <...> in path, to make a Flask friendly path
                        data.path = data.path.replace('{', '<').replace('}', '>')

                    if data.param_in_body and data.param_in_query:
                        raise Exception("Does not support params in both body and param (%s %s)" % (method, path))

                    if data.param_in_body and data.param_in_formdata:
                        raise Exception("Does not support params in both body and formdata (%s %s)" % (method, path))

                else:
                    data.no_params = True

                callback(data)



import imp
import os
import pprint
import json
import responses
from mock import patch, MagicMock
from pymacaron_core.swagger.client import _format_flask_url
from pymacaron_core.exceptions import PyMacaronCoreException, ValidationError
from pymacaron_core.models import get_model


utils = imp.load_source('common', os.path.join(os.path.dirname(__file__), 'utils.py'))


class Test(utils.PymTest):


    def setUp(self):
        super(Test, self).setUp()


    @responses.activate
    def test_client_with_query_param(self):
        handler, _ = self.generate_client_and_spec(self.yaml_query_param)

        responses.add(
            responses.GET, "http://some.server.com:80/v1/some/path",
            body=json.dumps({"foo": "a", "bar": "b"}),
            status=200,
            content_type="application/json"
        )

        res = handler(arg1='this', arg2='that')

        print("response: " + pprint.pformat(res))
        self.assertEqual(type(res).__name__, 'Result')
        self.assertEqual(res.foo, 'a')
        self.assertEqual(res.bar, 'b')


    @patch('pymacaron_core.swagger.client.requests')
    def test_requests_parameters_with_query_param(self, requests):
        requests.get = MagicMock()
        handler, _ = self.generate_client_and_spec(self.yaml_query_param)

        with self.assertRaises(PyMacaronCoreException):
            handler(arg1='this', arg2='that')
        # self.assertEqual("Expected 1 caller, got 0", str(e.exception))

        requests.get.assert_called_once_with(
            'http://some.server.com:80/v1/some/path',
            data=None,
            headers={'Content-Type': 'application/json'},
            params={'arg1': 'this', 'arg2': 'that'},
            timeout=(10, 10),
            verify=True
        )


    @responses.activate
    def test_client_with_body_param(self):
        handler, spec = self.generate_client_and_spec(self.yaml_body_param)

        responses.add(
            responses.POST,
            "http://some.server.com:80/v1/some/path",
            body=json.dumps({"foo": "a", "bar": "b"}),
            status=200,
            content_type="application/json"
        )

        # Only 1 parameter expected
        with self.assertRaises(ValidationError):
            res = handler()
        with self.assertRaises(ValidationError):
            res = handler(1, 2)

        # Send a valid parameter object
        model_class = get_model('Param')
        param = model_class(arg1='a', arg2='b')

        res = handler(param)
        self.assertEqual(type(res).__name__, 'Result')
        self.assertEqual(res.foo, 'a')
        self.assertEqual(res.bar, 'b')


    @patch('pymacaron_core.swagger.client.requests')
    def test_requests_parameters_with_body_param(self, requests):
        handler, spec = self.generate_client_and_spec(self.yaml_body_param)
        model_class = get_model('Param')
        param = model_class(arg1='a', arg2='b')

        with self.assertRaises(PyMacaronCoreException):
            handler(param)

        requests.post.assert_called_once_with(
            'http://some.server.com:80/v1/some/path',
            data=json.dumps({"arg1": "a", "arg2": "b"}),
            headers={'Content-Type': 'application/json'},
            params=None,
            timeout=(10, 10),
            verify=True
        )


# def test_client_with_auth_required():
#     pass


    def test__format_flask_url(self):
        ref = {
            'item_id': '1234',
            'path': 'abcd',
        }

        data = ref.copy()
        u = _format_flask_url(
            "/v1/seller/<item_id>/<path>/foo",
            data
        )
        self.assertEqual(u, "/v1/seller/1234/abcd/foo", u)
        self.assertEqual(len(list(data.keys())), 0)

        data = ref.copy()
        u = _format_flask_url(
            "/v1/seller/<item_id>/<path>/foo/<item_id>",
            data,
        )
        self.assertEqual(u, "/v1/seller/1234/abcd/foo/1234", u)
        self.assertEqual(len(list(data.keys())), 0)

        data = ref.copy()
        u = _format_flask_url(
            "/v1/seller/<item_id>/foo",
            data,
        )
        self.assertEqual(u, "/v1/seller/1234/foo", u)
        self.assertEqual(len(list(data.keys())), 1)
        self.assertEqual(data['path'], 'abcd')



    @responses.activate
    def test_client_with_path_param(self):
        handler, spec = self.generate_client_and_spec(self.yaml_path_param)

        responses.add(
            responses.GET,
            "http://some.server.com:80/v1/some/123/path/456",
            body=json.dumps({"foo": "a", "bar": "b"}),
            status=200,
            content_type="application/json"
        )

        # Make a valid call
        res = handler(foo=123, bar=456)
        self.assertEqual(type(res).__name__, 'Result')
        self.assertEqual(res.foo, 'a')
        self.assertEqual(res.bar, 'b')


    @patch('pymacaron_core.swagger.client.requests')
    def test_requests_parameters_with_path_params(self, requests):
        handler, spec = self.generate_client_and_spec(self.yaml_path_param)

        with self.assertRaises(PyMacaronCoreException):
            handler(foo=123, bar=456)

        requests.get.assert_called_once_with(
            'http://some.server.com:80/v1/some/123/path/456',
            data=None,
            headers={'Content-Type': 'application/json'},
            params=None,
            timeout=(10, 10),
            verify=True)


    @patch('pymacaron_core.swagger.client.requests')
    def test_handler_extra_parameters(self, requests):
        handler, spec = self.generate_client_and_spec(self.yaml_path_param)

        with self.assertRaises(PyMacaronCoreException):
            handler(
                foo=123,
                bar=456,
                max_attempts=2,
                read_timeout=6,
                connect_timeout=8
            )

        requests.get.assert_called_once_with(
            'http://some.server.com:80/v1/some/123/path/456',
            data=None,
            headers={'Content-Type': 'application/json'},
            params=None,
            timeout=(8, 6),
            verify=True
        )


    @responses.activate
    def test_client_with_path_query_param(self):
        handler, spec = self.generate_client_and_spec(self.yaml_path_query_param)

        responses.add(
            responses.GET,
            "http://some.server.com:80/v1/some/123/path",
            body=json.dumps({"foo": "a", "bar": "b"}),
            status=200,
            content_type="application/json"
        )

        # Make a valid call
        res = handler(foo=123, bar=456)
        self.assertEqual(type(res).__name__, 'Result')
        self.assertEqual(res.foo, 'a')
        self.assertEqual(res.bar, 'b')


    @patch('pymacaron_core.swagger.client.requests')
    def test_requests_parameters_with_path_query_params(self, requests):
        handler, spec = self.generate_client_and_spec(self.yaml_path_query_param)

        with self.assertRaises(PyMacaronCoreException):
            handler(foo=123, bar=456)

        requests.get.assert_called_once_with(
            'http://some.server.com:80/v1/some/123/path',
            data=None,
            headers={'Content-Type': 'application/json'},
            params={'bar': 456},
            timeout=(10, 10),
            verify=True)



    @responses.activate
    def test_client_with_path_body_param(self):
        handler, spec = self.generate_client_and_spec(self.yaml_path_body_param)

        responses.add(
            responses.GET,
            "http://some.server.com:80/v1/some/123/path",
            body=json.dumps({"foo": "a", "bar": "b"}),
            status=200,
            content_type="application/json"
        )

        # Send a valid parameter object
        model_class = get_model('Param')
        param = model_class(arg1='a', arg2='b')

        res = handler(param, foo=123)
        self.assertEqual(type(res).__name__, 'Result')
        self.assertEqual(res.foo, 'a')
        self.assertEqual(res.bar, 'b')

        # Only 1 parameter expected
        with self.assertRaises(ValidationError) as e:
            res = handler(foo=123)
        self.assertTrue('expects exactly' in str(e.exception))

        with self.assertRaises(ValidationError) as e:
            res = handler(1, 2, foo=123)
        self.assertTrue('expects exactly' in str(e.exception))

        with self.assertRaises(ValidationError) as e:
            res = handler(param)
        self.assertTrue('Missing some arguments' in str(e.exception))


    @patch('pymacaron_core.swagger.client.requests')
    def test_requests_parameters_with_path_body_params(self, requests):
        handler, spec = self.generate_client_and_spec(self.yaml_path_body_param)

        model_class = get_model('Param')
        param = model_class(arg1='a', arg2='b')

        with self.assertRaises(PyMacaronCoreException):
            handler(param, foo=123)

        requests.get.assert_called_once_with(
            'http://some.server.com:80/v1/some/123/path',
            data=json.dumps({"arg1": "a", "arg2": "b"}),
            headers={'Content-Type': 'application/json'},
            params=None,
            timeout=(10, 10),
            verify=True
        )


    @responses.activate
    def test_client_unknown_method(self):
        y = self.yaml_query_param
        y = y.replace('get:', 'foobar:')

        with self.assertRaises(PyMacaronCoreException) as e:
            handler, spec = self.generate_client_and_spec(y)
        self.assertTrue('BUG: method FOOBAR for /v1/some/path is not supported' in str(e.exception))


    @patch('pymacaron_core.swagger.client.requests')
    def test_requests_client_override_read_timeout(self, requests):
        handler, spec = self.generate_client_and_spec(self.yaml_path_query_param)

        with self.assertRaises(PyMacaronCoreException):
            handler(read_timeout=50, foo='123', bar='456')

        requests.get.assert_called_once_with(
            'http://some.server.com:80/v1/some/123/path',
            data=None,
            headers={'Content-Type': 'application/json'},
            params={'bar': '456'},
            timeout=(10, 50),
            verify=True)


    @patch('pymacaron_core.swagger.client.requests')
    def test_requests_client_override_connect_timeout(self, requests):
        handler, spec = self.generate_client_and_spec(self.yaml_path_query_param)

        with self.assertRaises(PyMacaronCoreException):
            handler(connect_timeout=50, foo='123', bar='456')

        requests.get.assert_called_once_with(
            'http://some.server.com:80/v1/some/123/path',
            data=None,
            headers={'Content-Type': 'application/json'},
            params={'bar': '456'},
            timeout=(50, 10),
            verify=True)


    @responses.activate
    def test_client_error_callback_return_dict(self):

        def callback(e):
            return {'error': str(e)}

        handler, spec = self.generate_client_and_spec(
            self.yaml_path_body_param,
            callback=callback,
        )

        responses.add(
            responses.GET,
            "http://some.server.com:80/v1/some/123/path",
            body=json.dumps({"foo": "a", "bar": "b"}),
            status=200,
            content_type="application/json"
        )

        # Send a valid parameter object
        model_class = get_model('Param')
        param = model_class(arg1='a', arg2='b')

        res = handler(param)
        print("got: %s" % res)
        self.assertDictEqual(
            res,
            {
                'error': 'Missing some arguments to format url: http://some.server.com:80/v1/some/<foo>/path'
            }
        )


# TODO: test max_attempts?



import logging
from werkzeug import FileStorage
from bravado_core.request import IncomingRequest


log = logging.getLogger(__name__)


class FlaskRequestProxy(IncomingRequest):
    """Take a flask.request object and make it look like a
    bravado_core.request.IncomingRequest"""

    path = None
    query = None
    form = None
    headers = None
    _json = None

    def __init__(self, request, has_data):
        self.request = request
        self.query = request.args
        self.path = request.view_args
        self.headers = request.headers
        self.files = {}
        self._json = {}

        if has_data:
            # has_data == True means there are key-value parameters we can extract from the request body

            # If the request contained no data, no need to analyze it further
            if len(self.request.get_data()) == 0:
                return

            # Let's try to convert whatever content-type we got in the request to something json-like
            ctype = self.request.content_type
            if not ctype:
                # If no content-type specified, assume json
                ctype = 'application/json'

            if ctype.startswith('application/x-www-form-urlencoded'):
                # Store the request's form
                self._json = self.request.form.to_dict()

            elif ctype.startswith('multipart/form-data'):
                # Store the request's form and files
                self._json = self.request.form.to_dict()

                # Go through all the objects passed in form-data and try converting to something json-friendly
                files = self.request.files.to_dict()
                for k in list(files.keys()):
                    v = files[k]
                    if type(v) is FileStorage:
                        # In bravado_core.request.IncomingRequest, 'files' contains a dict of param name to content
                        name = v.name
                        self.files[name] = v.read()
                        # Since bravado drops the filename and mimetype, we explicitely add them to the files dict
                        self.files['%s_filename' % name] = v.filename
                        self.files['%s_mimetype' % name] = v.content_type
                    else:
                        raise Exception("Support for multipart/form-data containing %s is not implemented" % type(v))
            else:
                # Assuming we got a json body
                self._json = self.request.get_json(force=True)

    def json(self):
        # Convert a weltkreuz ImmutableDict to a simple python dict
        return self._json



import logging
from datetime import datetime
from copy import deepcopy
from bravado_core.marshal import marshal_schema_object
from bravado_core.unmarshal import unmarshal_model
import bravado_core.model
from pymacaron_core.exceptions import ValidationError
from pymacaron_core.utils import get_function


log = logging.getLogger(__name__)


class Models():
    """Class holding all generated models"""
    pass


def get_model(model_name):
    if hasattr(Models, model_name):
        return getattr(Models, model_name)
    raise ValidationError("Swagger spec has no definition for model %s" % model_name)


class PyMacaronModel(object):
    """Instances of PyMacaron Model are passed to and returned by the API
    endpoints.

    They encapsulate an instance of a Bravado model that holds the instance's
    attributes, they may inherit from a parent class, they can be serialized to
    and from json, they can be loaded from and to a persistent database.

    """

    # Class variables: Name of this model and pymacaron-core ApiSpec of that model
    __model_name = None
    __swagger_dict = None
    __swagger_spec = None

    #
    # Delegate getter/setter/etc to Bravado model
    #

    def __setattr__(self, k, v):
        if k in getattr(self, '__property_names'):
            setattr(getattr(self, '__bravado_instance'), k, v)
        else:
            super().__setattr__(k, v)


    def __getattr__(self, k):
        # NOTE: overriding __getattr__ is tricky since we need __property_names early,
        # and getattr() is used by hasattr() to check if an attribute exists
        if k.endswith('__property_names'):
            return getattr(self, '__property_names')
        elif k in getattr(self, '__property_names'):
            return getattr(getattr(self, '__bravado_instance'), k)
        elif k not in dir(self):
            raise AttributeError("Model '%s' has no attribute %s" % (getattr(self, '__model_name'), k))
        else:
            return getattr(self, k)


    def __delattr__(self, k):
        if k in getattr(self, '__property_names'):
            delattr(getattr(self, '__bravado_instance'), k)
        else:
            super().__delattr__(k)


    def __getitem__(self, k):
        if k not in getattr(self, '__property_names'):
            raise AttributeError("Model '%s' has no attribute %s" % (getattr(self, '__model_name'), k))
        return getattr(getattr(self, '__bravado_instance'), k)


    def __setitem__(self, k, v):
        if k not in getattr(self, '__property_names'):
            raise AttributeError("Model '%s' has no attribute %s" % (getattr(self, '__model_name'), k))
        setattr(getattr(self, '__bravado_instance'), k, v)


    def __delitem__(self, k):
        if k not in getattr(self, '__property_names'):
            raise AttributeError("Model '%s' has no attribute %s" % (getattr(self, '__model_name'), k))
        delattr(getattr(self, '__bravado_instance'), k)


    def __eq__(self, other):
        if type(self) is not type(other):
            return False
        return getattr(self, '__bravado_instance') == getattr(other, '__bravado_instance')


    def __repr__(self):
        return 'PyMacaron:%s:%s' % (getattr(self, '__model_name'), str(getattr(self, '__bravado_instance')))


    #
    # Convenience methods
    #


    def update_from_dict(self, d, ignore_none=False):
        """Take a dictionary of key-values representing attribute and values to update
        in the encapsulated Bravado instance. Note that setting an attribute to
        None actually means removing it from the model instance, and thereby
        from its json representation, except if ignore_none is True, in which case
        the attribute is kept unchanged.
        """

        for k, v in d.items():
            if v is None and ignore_none:
                pass
            elif v is None:
                delattr(getattr(self, '__bravado_instance'), k)
            else:
                setattr(getattr(self, '__bravado_instance'), k, v)


    def clone(self):
        """Return a clone of self"""
        j = self.to_json()
        return self.__class__.from_json(j)

    #
    # JSON marshal/unmarshal
    #


    def to_json(self, keep_datetime=False):
        """Return a json representation of this PyMacaron object - If keep_datetime is set,
        will keep attributes that are datetime unchanged.
        """
        log.debug("Marshalling %s into json" % getattr(self, '__model_name'))
        datetimes = {}
        if keep_datetime:
            for k in self.__property_names:
                if hasattr(self, k) and type(getattr(self, k)) is datetime:
                    datetimes[k] = getattr(self, k)
        j = marshal_schema_object(
            getattr(self, '__swagger_spec'),
            getattr(self, '__swagger_dict'),
            self.to_bravado(),
        )
        if datetimes:
            j.update(datetimes)
        return j


    @classmethod
    def from_json(cls, j, keep_datetime=False):
        """Take a json dictionary and return a model instance"""
        log.debug("Unmarshalling json into %s" % getattr(cls, '__model_name'))
        datetimes = {}
        if keep_datetime:
            for k in list(j.keys()):
                if type(j[k]) is datetime:
                    datetimes[k] = j[k]
                    del j[k]

        m = unmarshal_model(
            getattr(cls, '__swagger_spec'),
            getattr(cls, '__swagger_dict'),
            j
        )

        if datetimes:
            for k in datetimes:
                setattr(m, k, datetimes[k])

        return cls.from_bravado(m)


    def get_model_name(self):
        """Return the name of the OpenAPI schema object describing this PyMacaron Model instance"""
        return getattr(self, '__model_name')

    #
    # Methods to cast a PyMacaron Model to/from a Bravado Model
    #

    def to_bravado(self):
        """Return a pure Bravado Model representing self"""

        # Get internal Bravado instance and clone it
        o = getattr(self, '__bravado_instance')
        o = deepcopy(o)

        # Do the same recursively with every nested PyMacaron Model
        for k in getattr(self, '__property_names'):
            v = getattr(o, k)
            if isinstance(v, PyMacaronModel):
                setattr(o, k, v.to_bravado())
            elif type(v) is list:
                for i in range(len(v)):
                    if isinstance(v[i], PyMacaronModel):
                        v[i] = v[i].to_bravado()
        return o


    @classmethod
    def from_bravado(cls, o):
        """Take a bravado Model instance and return a PyMacaron Model instance"""

        # Clone bravado instance and inject it into a matching PyMacaron model instance
        o = deepcopy(o)
        p = cls()
        setattr(p, '__bravado_instance', o)

        # Now cast from bravado to pymacaron models all the attributes of this model
        for k in getattr(p, '__property_names'):
            v = getattr(o, k)
            if isinstance(v, bravado_core.model.Model):
                cls = get_model(v.__class__.__name__)
                setattr(o, k, cls.from_bravado(v))
            elif type(v) is list:
                for i in range(len(v)):
                    if isinstance(v[i], bravado_core.model.Model):
                        cls = get_model(v[i].__class__.__name__)
                        v[i] = cls.from_bravado(v[i])
        return p


def generate_model_class(name=None, bravado_class=None, swagger_dict=None, swagger_spec=None, parent_name=None, persist=None, properties={}):
    """Dynamically generate a pymacaron.models.<model_name> class able to
    instantiate that model.

    :name: the model name, as in the swagger spec
    :parent_name: complete name (module path + class name) of a class that this model should inherit from.
    :param persist: name of a package or class that implements the 'load_from_db' and 'save_to_db' methods.
    """

    if parent_name:
        assert type(parent_name) is str

    if persist:
        assert type(persist) is str

    # Which parents are we inheriting?
    parents = (PyMacaronModel, )
    if parent_name:
        parent_class = get_function(parent_name)
        parents = parents + (parent_class, )

    # Is this model persistent?
    persistence_class = None
    if persist:
        persistence_class = get_function(persist)
        parents = parents + (persistence_class, )

    # Generate the instance's constructor
    def init(self, *args, **kwargs):
        self.__bravado_instance = bravado_class(*args, **kwargs)

    # And generate the model's class
    o = type(
        name,
        parents,
        {
            '__init__': init,
            '__model_name': name,
            '__persistence_class__': persist,
            '__property_names': list(properties.keys()),
            '__swagger_spec': swagger_spec,
            '__swagger_dict': swagger_dict,
        },
    )

    # And remember the mapping between this bravado model and its pymacaron model
    setattr(Models, name, o)

    return o



import requests
import pprint
import jsonschema
import json
import logging
import time
import urllib.request
import urllib.parse
import urllib.error
from requests.exceptions import ReadTimeout, ConnectTimeout
from pymacaron_core.exceptions import PyMacaronCoreException, ValidationError
from pymacaron_core.utils import get_function
from bravado_core.response import unmarshal_response


log = logging.getLogger(__name__)


try:
    from flask import _app_ctx_stack as stack
except ImportError:
    from flask import _request_ctx_stack as stack


def generate_client_callers(spec, timeout, error_callback, local, app):
    """Return a dict mapping method names to anonymous functions that
    will call the server's endpoint of the corresponding name as
    described in the api defined by the swagger dict and bravado spec"""

    callers_dict = {}

    def mycallback(endpoint):
        if not endpoint.handler_client:
            return

        callers_dict[endpoint.handler_client] = _generate_client_caller(spec, endpoint, timeout, error_callback, local, app)

    spec.call_on_each_endpoint(mycallback)

    return callers_dict


def _generate_request_arguments(url, spec, endpoint, headers, args, kwargs):
    # Prepare (g)requests arguments
    data = None
    params = None
    custom_url = url

    if hasattr(stack.top, 'call_id'):
        headers['PymCallID'] = stack.top.call_id
    if hasattr(stack.top, 'call_path'):
        headers['PymCallPath'] = stack.top.call_path

    if endpoint.param_in_path:
        # Fill url with values from kwargs, and remove those params from kwargs
        custom_url = _format_flask_url(url, kwargs)

    if endpoint.param_in_query:
        # The query parameters are contained in **kwargs
        params = kwargs
        # TODO: validate params? or let the server do that...
    elif endpoint.param_in_body:
        # The body parameter is the first elem in *args
        if len(args) != 1:
            raise ValidationError("%s expects exactly 1 parameter" % endpoint.handler_client)
        data = json.dumps(spec.model_to_json(args[0]))

    # Prune undefined parameters that would otherwise be turned into '=None'
    # query params
    if params:
        for k in list(params.keys()):
            if params[k] is None:
                del params[k]

    return custom_url, params, data, headers


def _generate_client_caller(spec, endpoint, timeout, error_callback, local, app):

    if local:
        assert app

    # Is the endpoint available locally?
    if local:
        url = endpoint.path.lstrip('/')
    else:
        url = "%s://%s:%s/%s" % (
            spec.protocol,
            spec.host,
            spec.port,
            endpoint.path.lstrip('/')
        )

    # Get eventual decorator and http method
    decorator = None
    if endpoint.decorate_request:
        decorator = get_function(endpoint.decorate_request)

    method = endpoint.method.lower()
    if method not in ('get', 'post', 'patch', 'put', 'delete'):
        raise PyMacaronCoreException("BUG: method %s for %s is not supported. Only get and post are." %
                                     (endpoint.method, endpoint.path))

    # Are we doing a local call?
    if local:
        def local_client(*args, **kwargs):
            """Just call the local method"""
            log.info("Calling %s locally via flask test_client" % (endpoint.path))

            headers = {'Content-Type': 'application/json'}
            headers.update(kwargs.get('request_headers', {}))

            # Remove magic client parameters before passing on
            for k in ('max_attempts', 'read_timeout', 'connect_timeout', 'request_headers'):
                if k in kwargs:
                    del kwargs[k]

            custom_url, params, data, headers = _generate_request_arguments(url, spec, endpoint, headers, args, kwargs)
            if '<' in custom_url:
                # Some arguments were missing
                return error_callback(ValidationError("Missing some arguments to format url: %s" % custom_url))

            if params:
                for k, v in params.items():
                    if isinstance(v, str):
                        params[k] = v.encode('utf-8')
                custom_url = custom_url + '?' + urllib.parse.urlencode(params)
            log.info("Calling with params [%s]" % params)

            with app.test_client() as c:
                requests_method = getattr(c, method)
                if decorator:
                    requests_method = decorator(requests_method)

                response = requests_method(
                    custom_url,
                    data=data,
                    headers=headers
                )

            return response_to_result(response, method, custom_url, endpoint.operation, error_callback)

        return local_client

    # Else call over HTTP/HTTPS
    requests_method = getattr(requests, method)
    if decorator:
        requests_method = decorator(requests_method)

    def client(*args, **kwargs):
        """Call the server endpoint and handle marshaling/unmarshaling of parameters/result.

        client takes either a dict of query parameters, or an object representing the unique
        body parameter.
        """

        # Extract custom parameters from **kwargs
        headers = {'Content-Type': 'application/json'}
        max_attempts = 3
        read_timeout = timeout
        connect_timeout = timeout

        if 'max_attempts' in kwargs:
            max_attempts = kwargs['max_attempts']
            del kwargs['max_attempts']
        if 'read_timeout' in kwargs:
            read_timeout = kwargs['read_timeout']
            del kwargs['read_timeout']
        if 'connect_timeout' in kwargs:
            connect_timeout = kwargs['connect_timeout']
            del kwargs['connect_timeout']
        if 'request_headers' in kwargs:
            headers.update(kwargs['request_headers'])
            del kwargs['request_headers']

        custom_url, params, data, headers = _generate_request_arguments(url, spec, endpoint, headers, args, kwargs)

        if '<' in custom_url:
            # Some arguments were missing
            return error_callback(ValidationError("Missing some arguments to format url: %s" % custom_url))

        # TODO: refactor this left-over from the time of async/grequests support and simplify!
        return ClientCaller(requests_method, custom_url, data, params, headers, read_timeout, connect_timeout, endpoint.operation, endpoint.method, error_callback, max_attempts, spec.verify_ssl).call()

    return client


def _format_flask_url(url, params):
    # TODO: make this code more robust: error if some params are left unmatched
    # or if url still contains placeholders after replacing
    remove = []
    for name, value in params.items():
        if "<%s>" % name in url:
            url = url.replace("<%s>" % name, str(value))
            remove.append(name)

    for name in remove:
        del params[name]

    return url


def response_to_result(response, method, url, operation, error_callback):

    # Monkey patching flask test_client response if necessary
    if not hasattr(response, 'text'):
        data = response.data.decode("utf-8")
        setattr(response, 'text', data)
        j = json.loads(data)

        def get_json():
            return j

        setattr(response, 'json', get_json)

    # If the remote-server returned an error, raise it as a local PyMacaronCoreException
    if str(response.status_code) != '200':
        log.warn("Call to %s %s returns error: %s" % (method, url, response.text))
        if 'error_description' in response.text:
            # We got a PyMacaronCoreException: unmarshal it and return as valid
            # return value UGLY FRAGILE CODE. To be replaced by proper
            # exception scheme
            pass
        else:
            # Unknown exception...
            log.info("Unknown exception: " + response.text)
            k = PyMacaronCoreException("Call to %s %s returned unknown exception: %s" % (method, url, response.text))
            k.status_code = response.status_code
            c = error_callback
            if hasattr(c, '__func__'):
                c = c.__func__
            return c(k)

    # Now transform the request's Response object into an instance of a
    # swagger model
    try:
        result = unmarshal_response(response, operation)
    except jsonschema.exceptions.ValidationError as e:
        log.warn("Failed to unmarshal response: %s" % e)
        k = ValidationError("Failed to unmarshal response because: %s" % str(e))
        c = error_callback
        if hasattr(c, '__func__'):
            c = c.__func__
        return c(k)

    log.info("Call to %s %s returned an instance of %s" % (method, url, type(result)))
    return result


class ClientCaller():

    def __init__(self, requests_method, url, data, params, headers, read_timeout, connect_timeout, operation, method, error_callback, max_attempts, verify_ssl):
        assert max_attempts >= 1
        self.requests_method = requests_method
        self.url = url
        self.operation = operation
        self.data = data
        self.params = params
        self.headers = headers
        self.read_timeout = read_timeout
        self.connect_timeout = connect_timeout
        self.method = method.upper()
        self.error_callback = error_callback
        self.max_attempts = max_attempts
        self.verify_ssl = verify_ssl

    def _method_is_safe_to_retry(self):
        return self.method in ('GET', 'PATCH')

    def _call_retry(self, force_retry):
        """Call request and retry up to max_attempts times (or none if self.max_attempts=1)"""
        last_exception = None
        for i in range(self.max_attempts):
            try:
                log.info("Calling %s %s" % (self.method, self.url))
                response = self.requests_method(
                    self.url,
                    data=self.data,
                    params=self.params,
                    headers=self.headers,
                    timeout=(self.connect_timeout, self.read_timeout),
                    verify=self.verify_ssl,
                )

                if response is None:
                    log.warn("Got response None")
                    if self._method_is_safe_to_retry():
                        delay = 0.5 + i * 0.5
                        log.info("Waiting %s sec and Retrying since call is a %s" % (delay, self.method))
                        time.sleep(delay)
                        continue
                    else:
                        raise PyMacaronCoreException("Call %s %s returned empty response" % (self.method, self.url))

                return response

            except Exception as e:

                last_exception = e

                retry = force_retry

                if isinstance(e, ReadTimeout):
                    # Log enough to help debugging...
                    log.warn("Got a ReadTimeout calling %s %s" % (self.method, self.url))
                    log.warn("Exception was: %s" % str(e))
                    resp = e.response
                    if not resp:
                        log.info("Requests error has no response.")
                        # TODO: retry=True? Is it really safe?
                    else:
                        b = resp.content
                        log.info("Requests has a response with content: " + pprint.pformat(b))
                    if self._method_is_safe_to_retry():
                        # It is safe to retry
                        log.info("Retrying since call is a %s" % self.method)
                        retry = True

                elif isinstance(e, ConnectTimeout):
                    log.warn("Got a ConnectTimeout calling %s %s" % (self.method, self.url))
                    log.warn("Exception was: %s" % str(e))
                    # ConnectTimeouts are safe to retry whatever the call...
                    retry = True

                if retry:
                    continue
                else:
                    raise e

        # max_attempts has been reached: propagate the last received Exception
        if not last_exception:
            last_exception = Exception("Reached max-attempts (%s). Giving up calling %s %s" % (self.max_attempts, self.method, self.url))
        raise last_exception

    def call(self, force_retry=False):
        response = self._call_retry(force_retry)
        return response_to_result(response, self.method, self.url, self.operation, self.error_callback)



import unittest
from pymacaron_core.swagger.api import API
from pymacaron_core.models import get_model
from pymacaron_core.models import PyMacaronModel


#
# Swagger spec
#

yaml_str = """
swagger: '2.0'
info:
  version: '0.0.1'
host: some.server.com
schemes:
  - http
produces:
  - application/json
definitions:

  Foo:
    type: object
    properties:
      s:
        type: string
      i:
        type: integer
      o:
        $ref: '#/definitions/Bar'
      lst:
        type: array
        items:
          type: string
      lo:
        type: array
        items:
          $ref: '#/definitions/Bar'

  Bar:
    type: object
    properties:
      s:
        type: string
      o:
        $ref: '#/definitions/Baz'

  Baz:
    type: object
    properties:
      s:
        type: string

"""

#
# Tests
#

class Tests(unittest.TestCase):

    def setUp(self):
        API('somename', yaml_str=yaml_str)


    def test__setattr__getattr(self):
        o = get_model('Foo')()

        # set/get a bravado attribute
        self.assertEqual(o.s, None)
        self.assertEqual(getattr(o, 's'), None)

        o.s = 'bob'
        self.assertEqual(o.s, 'bob')
        self.assertEqual(getattr(o, 's'), 'bob')

        # Make sure it's really the Bravado instance's attribute that was updated
        self.assertTrue('s' not in dir(o))
        self.assertEqual(getattr(o, '__bravado_instance').s, 'bob')

        o.s = None
        self.assertEqual(o.s, None)
        self.assertEqual(getattr(o, 's'), None)

        setattr(o, 's', 'bob')
        self.assertTrue('s' not in dir(o))
        self.assertEqual(getattr(o, '__bravado_instance').s, 'bob')
        self.assertEqual(o.s, 'bob')
        self.assertEqual(getattr(o, 's'), 'bob')

        setattr(o, 's', None)
        self.assertTrue('s' not in dir(o))
        self.assertEqual(getattr(o, '__bravado_instance').s, None)
        self.assertEqual(o.s, None)
        self.assertEqual(getattr(o, 's'), None)

        # set/get a local attribute
        with self.assertRaises(Exception) as context:
            o.local
        self.assertTrue("Model 'Foo' has no attribute local" in str(context.exception))

        with self.assertRaises(Exception) as context:
            getattr(o, 'local')
        self.assertTrue("Model 'Foo' has no attribute local" in str(context.exception))

        o.local = 'bob'
        self.assertTrue('local' in dir(o))
        self.assertEqual(o.local, 'bob')
        self.assertEqual(getattr(o, 'local'), 'bob')

        o.local = None
        self.assertEqual(o.local, None)
        self.assertEqual(getattr(o, 'local'), None)

        setattr(o, 'local', 'bob')
        self.assertEqual(o.local, 'bob')
        self.assertEqual(getattr(o, 'local'), 'bob')

        setattr(o, 'local', None)
        self.assertEqual(o.local, None)
        self.assertEqual(getattr(o, 'local'), None)


    def test__hasattr(self):
        o = get_model('Foo')()

        self.assertTrue(hasattr(o, 's'))
        self.assertFalse(hasattr(o, 'local'))

        o.local = None
        self.assertTrue(hasattr(o, 'local'))


    def test__delattr(self):
        o = get_model('Foo')()

        o.s = 'bob'
        self.assertEqual(o.s, 'bob')
        del o.s
        self.assertTrue(hasattr(o, 's'))
        self.assertEqual(o.s, None)

        o.s = 'bob'
        self.assertEqual(o.s, 'bob')
        delattr(o, 's')
        self.assertTrue(hasattr(o, 's'))
        self.assertEqual(o.s, None)

        o.local = 'bob'
        self.assertEqual(o.local, 'bob')
        del o.local
        self.assertFalse(hasattr(o, 'local'))
        with self.assertRaises(Exception) as context:
            o.local
        self.assertTrue("Model 'Foo' has no attribute local" in str(context.exception))

        o.local = 'bob'
        self.assertEqual(o.local, 'bob')
        delattr(o, 'local')
        self.assertFalse(hasattr(o, 'local'))
        with self.assertRaises(Exception) as context:
            o.local
        self.assertTrue("Model 'Foo' has no attribute local" in str(context.exception))


    def test__getitem__setitem__delitem(self):
        o = get_model('Foo')()
        self.assertEqual(o.s, None)
        self.assertEqual(o['s'], None)

        o['s'] = 'bob'
        self.assertEqual(o.s, 'bob')
        self.assertEqual(o['s'], 'bob')

        o['s'] = None
        self.assertEqual(o.s, None)
        self.assertEqual(o['s'], None)

        o['s'] = 'bob'
        self.assertEqual(o.s, 'bob')
        del o['s']
        self.assertEqual(o.s, None)

        # But local attributes may not be set this way
        with self.assertRaises(Exception) as context:
            o['local']
        self.assertTrue("Model 'Foo' has no attribute local" in str(context.exception))

        with self.assertRaises(Exception) as context:
            o['local'] = 123
        self.assertTrue("Model 'Foo' has no attribute local" in str(context.exception))

        with self.assertRaises(Exception) as context:
            del o['local']
        self.assertTrue("Model 'Foo' has no attribute local" in str(context.exception))


    def test__eq(self):
        Foo = get_model('Foo')
        Bar = get_model('Bar')
        a = Foo(s='abc', i=12, o=Bar(s='def'))
        b = Foo(s='abc', i=12, o=Bar(s='def'))

        self.assertEqual(a, b)
        self.assertNotEqual(a, 'bob')

        # Adding local parameters does not affect eq
        a.local = 'whatever'
        self.assertEqual(a, b)

        # Changing bravado values makes them different
        a.o.s = '123'
        self.assertNotEqual(a, b)


    def test__to_json__from_json(self):
        Foo = get_model('Foo')
        Bar = get_model('Bar')
        Baz = get_model('Baz')
        a = Foo(
            s='abc',
            i=12,
            lst=['a', 'b', 'c'],
            o=Bar(
                s='1',
                o=Baz(
                    s='2'
                )
            ),
            lo=[
                Baz(s='r'),
                Baz(s='t'),
                Baz(s='u'),
                Baz(),
            ]
        )
        self.assertTrue(isinstance(a, PyMacaronModel))

        j = a.to_json()
        self.assertEqual(
            j,
            {
                's': 'abc',
                'i': 12,
                'lst': ['a', 'b', 'c'],
                'o': {'o': {'s': '2'}, 's': '1'},
                'lo': [{'s': 'r'}, {'s': 't'}, {'s': 'u'}, {}],
            }
        )

        o = Foo.from_json(j)
        self.assertTrue(isinstance(o, PyMacaronModel))
        # TODO: o now has multiple attributes set to None, while a lacks them,
        # and bravado's __eq__ does not see None and absence as equal...
        # self.assertEqual(o, a)

        jj = o.to_json()
        self.assertEqual(jj, j)


    def test__update_from_dict(self):
        foo = get_model('Foo')()

        foo.update_from_dict({'s': 'bob'})
        self.assertEqual(
            foo.to_json(),
            {'s': 'bob'},
        )

        foo.update_from_dict({'s': 'abc', 'i': 12})
        self.assertEqual(
            foo.to_json(),
            {'s': 'abc', 'i': 12},
        )

        foo.update_from_dict({})
        self.assertEqual(
            foo.to_json(),
            {'s': 'abc', 'i': 12},
        )

        foo.update_from_dict({'i': None})
        self.assertEqual(
            foo.to_json(),
            {'s': 'abc'},
        )

        foo.update_from_dict({'s': None, 'i': 32}, ignore_none=True)
        self.assertEqual(
            foo.to_json(),
            {'s': 'abc', 'i': 32},
        )

        foo.update_from_dict({'s': None})
        self.assertEqual(
            foo.to_json(),
            {'i': 32},
        )
